import numpy as np
from .util import UKN

import pandas as pd
from sklearn.metrics import roc_auc_score
from .list_datapoints import ListofDatapoints
from .datapoint import Datapoint


def filter(listofdatapoints: ListofDatapoints, min_AUROC: int = 0.8):
    """Filters out duplicate sequences.
        Only keep the first occurence of a sequence if all the other structures are the same.

        Examples:
            >>> datapoints = ListofDatapoints([ Datapoint(reference='ref1', sequence='AACCGG', structure=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1]),\
                                Datapoint(reference='ref2', sequence='AACCGG', structure=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1]),\
                                Datapoint(reference='ref2', sequence='AACCGG', structure=[[1, 2], [3, 4]], dms=[1,0,0,0,0,1]),\
                                Datapoint(reference='ref4', sequence='AUGGC', structure=[[1, 2]], dms=[0,0,0,0,1]),\
                                Datapoint(reference='ref5', sequence='AUGGC', structure=[[0, 4], [2, 3]], dms=[0,0,0,0,0]),\
                                Datapoint(reference='ref6', sequence='not a regular sequence', structure=[[0, 4]], dms=[0,0,0,0,0]) ])
            >>> print(filter(datapoints, min_AUROC=0))
            Over a total of 6 datapoints, there are:
            ### OUTPUT
            - 3 valid datapoints
            ### MODIFIED
            - 1 multiple sequences with the same reference (renamed reference)
            - 1 duplicate sequences with different structure / dms / shape
            ### FILTERED OUT
            - 1 invalid datapoints (ex: sequence with non-regular characters)
            - 0 datapoints with bad structures
            - 2 duplicate sequences with the same structure / dms / shape
            - 0 datapoints removed because of low AUROC (<0)
        """

    # Remove None datapoints
    n_input_datapoints = len(listofdatapoints)
    datapoints, n_unvalid_datapoints = listofdatapoints.drop_none_dp()

    # Remove bad structures
    bad_structures_idx = []
    for idx, datapoint in enumerate(datapoints):
        if not datapoint._assert_structure():
            bad_structures_idx.append(idx)
    datapoints = [
        datapoint
        for idx, datapoint in enumerate(datapoints)
        if idx not in bad_structures_idx
    ]
    n_bad_structures_datapoints = len(bad_structures_idx)

    # Convert to pandas
    df = listofdatapoints.to_pandas(datapoints)

    # Remove duplicate or conflicting datapoints and keep track of the number of datapoints removed
    def drop_duplicates(df: pd.DataFrame, **kwargs):
        len_df_before = len(df)
        df.drop_duplicates(**kwargs)
        return len_df_before - len(df)

    # If multiple sequences with the same reference, rename the reference
    refs = dict()
    n_same_ref_datapoints = 0
    for idx, row in df.iterrows():
        if row["reference"] in refs:
            refs[row["reference"]] += 1
            df.at[idx, "reference"] = f"{row['reference']}_{refs[row['reference']]}"
            n_same_ref_datapoints += 1
        else:
            refs[row["reference"]] = 0

    # Keep only one datapoint per sequence and structure
    if "structure" in df.columns:
        n_duplicates_datapoints = drop_duplicates(
            df,
            subset=["sequence", "structure"],
            inplace=True,
            ignore_index=True,
            keep="first",
        )

    # Keep only one datapoint per sequence and signal
    for signal in ["dms", "shape"]:
        if signal in df.columns:
            if not "structure" in df.columns:
                n_duplicates_datapoints = 0
            n_duplicates_datapoints += drop_duplicates(
                df,
                subset=["sequence", signal],
                inplace=True,
                ignore_index=True,
                keep="first",
            )

    # Count how many multiple structures / dms with the same sequence
    n_same_seq_datapoints = 0
    for _, group in df.groupby("sequence"):
        if len(group) > 1:
            n_same_seq_datapoints += len(group) - 1

    ## Filter out references with low AUROC
    mask_high_AUROC = None
    if "structure" in df.columns and "dms" in df.columns or "shape" in df.columns:
        signal = "dms" if "dms" in df.columns else "shape"

        def calculate_auroc(row):
            # best signal is DMS, if unavailable use SHAPE
            signal = (
                "dms" if "dms" in row and not type(row["dms"]) == float else "shape"
            )
            if (
                (not "structure" in row)
                or (row["structure"] is None)
                or (type(row["structure"]) == float and np.isnan(row["structure"]))
                or len(row["structure"]) == 0
                or row['structure'] == [[]]
            ):
                return 1  #  if you can't calculate the AUROC, don't filter it out

            if row[signal] is None or (type(row[signal]) == float and np.isnan(row[signal])):
                return 1
            sig = np.array(row[signal])
            if signal == "dms":  # mask G/T/U bases for DMS
                mask_UKN = np.array([s in "GTU" for s in row["sequence"].upper()])
                sig[mask_UKN] = UKN

            isUnpaired = np.ones_like(sig)
            isUnpaired[np.array(row["structure"]).flatten()] = 0
            if set(isUnpaired[sig != UKN]) != set([0, 1]):
                return 0
            return roc_auc_score(isUnpaired[sig != UKN], sig[sig != UKN])

        # Create a boolean mask for rows with auroc score greater than or equal to a threshold
        mask_high_AUROC = df.apply(
            lambda row: calculate_auroc(row) >= min_AUROC, axis=1
        )
        df = df[mask_high_AUROC]

    # Convert back to list of datapoints
    listofdatapoints.datapoints = listofdatapoints.from_pandas(df)

    # Write report
    report = f"""Over a total of {n_input_datapoints} datapoints, there are:
### OUTPUT
- ALL: {len(listofdatapoints)} valid datapoints
- INCLUDED: {n_same_seq_datapoints} duplicate sequences with different structure / dms / shape
### MODIFIED
- {n_same_ref_datapoints} multiple sequences with the same reference (renamed reference)
### FILTERED OUT
- {n_unvalid_datapoints} invalid datapoints (ex: sequence with non-regular characters)
- {n_bad_structures_datapoints} datapoints with bad structures"""
    if "structure" in df.columns or "dms" in df.columns or "shape" in df.columns:
        report += f"""
- {n_duplicates_datapoints} duplicate sequences with the same structure / dms / shape"""
    if mask_high_AUROC is not None:
        report += f"""
- {np.sum(~mask_high_AUROC)} datapoints removed because of low AUROC (<{min_AUROC})"""

    return report



import os
import numpy as np

# Define the one-hot encodings for the sequences and structures
seq2int = {
    "X": 0,
    "A": 1,
    "C": 2,
    "G": 3,
    "U": 4,
}

int2seq = {v: k for k, v in seq2int.items()}

dot2int = {".": 1, "(": 2, ")": 3, "X": 0}
int2dot = {v: k for k, v in dot2int.items()}

UKN = -1000


class DreemUtils:
    def flatten_json(data):
        out, row = [], {}

        for k, v in data.copy().items():
            if type(v) != dict:
                row[k] = v
            else:
                row["reference"] = k
                for k2, v2 in v.items():
                    if type(v2) != dict:
                        row[k2] = v2
                    else:
                        row["section"] = k2
                        for k3, v3 in v2.items():
                            row["cluster"] = k3
                            if type(v3) != dict:
                                row[k3] = v3
                            else:
                                for k4, v4 in v3.items():
                                    row[k4] = v4
                                out.append(row.copy())
        return out

    def sort_dict(mut_profiles):
        sorting_key = lambda item: 1000 * {int: 0, str: 0, float: 0, list: 1, dict: 2}[
            type(item[1])
        ] + ord(item[0][0])
        mut_profiles = {k: mut_profiles[k] for k in sorted(mut_profiles)}
        mut_profiles = dict(sorted(mut_profiles.items(), key=sorting_key))
        for k, v in mut_profiles.items():
            if type(v) is not dict:
                continue
            mut_profiles[k] = DreemUtils.sort_dict(v)
        return mut_profiles


def standardize_sequence(sequence):
    sequence = sequence.upper()
    sequence = sequence.replace("T", "U")
    # trim the sequence
    sequence = sequence.replace(" ", "").replace("\n", "").replace("\t", "")
    return sequence


def sequence_has_regular_characters(sequence):
    return not (set(sequence) - set("ACGU"))


def dump_json(data, path):
    if os.path.dirname(path) != "":
        os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as f:
        f.write("{\n")
        for idx, (ref, attr) in enumerate(
            data.items()
        ):  # write the datapoints one by one to avoid memory issues
            remove_keys = []
            for k, v in attr.items():
                if (
                    v is None
                    or (type(v) == float and np.isnan(v))
                    # or (hasattr(v, "__iter__") and not len(v))
                ):
                    remove_keys.append(k)
                if isinstance(v, np.ndarray):
                    v = v.tolist()
            for k in remove_keys:
                del attr[k]
            f.write(str({ref: attr})[1:-1].replace("'", '"'))
            f.write(",\n" if idx != len(data) - 1 else "\n}")



from .parsers import *
from .util import (
    standardize_sequence,
    sequence_has_regular_characters,
)
import numpy as np
from .rnastructure import RNAstructure_singleton


class Datapoint:

    """A datapoint is a data structure corresponding to a sequence, a reference, and optionally a structure, dms and shape."""

    def __new__(cls, *args, **kwargs):
        # Check if the conditions are met before creating the object
        sequence = kwargs.get("sequence", args[0] if len(args) > 0 else None)
        reference = kwargs.get("reference", args[1] if len(args) > 1 else None)
        dotbracket = kwargs.get("dotbracket", None)
        if (
            sequence is None
            or reference is None
            or len(sequence) == 0
            or len(reference) == 0
        ):
            return None

        if dotbracket is not None:
            if not dotbracket.count("(") == dotbracket.count(")"):
                return None

        if not sequence_has_regular_characters(sequence):
            return None

        # If conditions are met, create and return the new instance
        instance = super().__new__(cls)
        return instance

    def __init__(
        self,
        sequence,
        reference,
        dotbracket=None,
        dms=None,
        shape=None,
        structure=None,
    ):
        for attr in [sequence, reference]:
            assert isinstance(
                attr, str
            ), f"Expected {attr} to be a string, got {type(attr)} instead."
            assert len(attr) > 0, f"Expected {attr} to be non-empty."

        # standardize the sequence
        sequence = standardize_sequence(sequence)

        if not sequence_has_regular_characters(sequence):
            raise Exception(
                f"Sequence {sequence} contains characters other than ACGTUacgtu."
            )

        self.reference = reference
        self.sequence = sequence
        self.structure = structure

        if structure is None and dotbracket is not None:
            self.structure = self.dotbracket_to_structure(dotbracket)
        if dms is not None:
            self.dms = self._format_signal(dms)
        if shape is not None:
            self.shape = self._format_signal(shape)

    def get_opt_dict(self):
        return {
            attr: eval(f"self.{attr}")
            for attr in ["structure", "dms", "shape"]
            if hasattr(self, attr)
        }

    def _assert_structure(self):
        """Ensures that the base pairs are 0=<bp<len(sequence) and that that there's maximum one pair per base.
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', dotbracket='((..))')
        >>> datapoint._assert_structure()
        True
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', dotbracket='((..)(')
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure=[[1, 2], [3, 4]])
        >>> datapoint._assert_structure()
        True
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure=[[1, 2], [3, 3]])
        >>>> datapoint._assert_structure()
        False
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure=[[1, 2], [3, 6]])
        >>> datapoint._assert_structure()
        False
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure=[[1, 2], [3, -1]])
        >>> datapoint._assert_structure()
        False
        """
        sequence = self.sequence
        structure = self.structure

        if structure is None or structure == [[]]:
            return True
        return (
            all(
                # check boundaries
                [
                    0 <= pair[0] < len(sequence)
                    and 0 <= pair[1] < len(sequence)
                    and pair[0] != pair[1]
                    for pair in structure
                ]
            )
            # check that there's maximum one pair per base
            and len(
                set([pair[0] for pair in structure] + [pair[1] for pair in structure])
            )
            == len(structure) * 2
        )

    def _assert_exists(self, attr):
        if not hasattr(self, attr):
            return False
        if getattr(self, attr) is None:
            return False
        if isinstance(getattr(self, attr), float) and np.isnan(getattr(self, attr)):
            return False
        return True

    def convert_arrays_to_tuple(self):
        if self._assert_exists("dms"):
            self.dms = tuple(self.dms)
        if self._assert_exists("shape"):
            self.shape = tuple(self.shape)
        if self._assert_exists("structure"):
            self.structure = tuple([tuple(pair) for pair in self.structure])

    def convert_arrays_to_list(self):
        if self._assert_exists("dms"):
            self.dms = list(self.dms)
        if self._assert_exists("shape"):
            self.shape = list(self.shape)
        if self._assert_exists("structure"):
            self.structure = [list(pair) for pair in self.structure]

    def _format_signal(self, signal):
        if not hasattr(signal, "__iter__"):  # handles none and nan
            return signal
        return [round(d, 4) for d in signal]

    def to_dict(self):
        return self.reference, {
            "sequence": self.sequence,
            **{k: v for k, v in self.get_opt_dict().items() if v is not None},
        }

    def to_flat_dict(self):
        return {
            "reference": self.reference,
            "sequence": self.sequence,
            **{k: v for k, v in self.get_opt_dict().items() if v is not None},
        }

    @classmethod
    def from_flat_dict(cls, d):
        return cls(
            sequence=d["sequence"],
            reference=d["reference"],
            **{k: v for k, v in d.items() if k not in ["sequence", "reference"]},
        )

    def __str__(self):
        return f'"{self.reference}":' + str(
            {
                "sequence": self.sequence,
                **{
                    k: v
                    for k, v in self.get_opt_dict().items()
                    if v is not None and not (type(v) == float and np.isnan(v))
                },
            }
        ).replace("'", '"').replace("(", "[").replace(")", "]").replace("None", "null")

    def __repr__(self) -> str:
        out = (
            f"{self.__class__.__name__}('{self.reference}', sequence='{self.sequence}'"
        )
        if hasattr(self, "structure"):
            out += f", structure={self.structure}"
        if hasattr(self, "dms"):
            out += f", dms={self.dms}"
        return out + ")"

    def dotbracket_to_structure(self, dotbracket):
        """Returns a list of tuples (i,j) where i and j are paired bases."""
        structure = set()
        stack = []
        for i, char in enumerate(dotbracket):
            if char == "(":
                stack.append(i)
            elif char == ")":
                structure.add((stack.pop(), i))
        return structure


class DatapointFactory:
    """Factory class to create datapoints from different formats."""

    def from_bpseq(bpseq_file):
        """Create a datapoint from a bpseq file. If predict_dms is True, the dms will be predicted using RNAstructure"""
        reference, sequence, structure = BPseq.parse(bpseq_file)
        sequence = standardize_sequence(sequence)

        if sequence_has_regular_characters(sequence):
            return Datapoint(
                sequence=sequence,
                reference=reference,
                structure=structure,
            )

    def from_ct(ct_file):
        """Create a datapoint from a ct file. If predict_dms is True, the dms will be predicted using RNAstructure"""
        reference, sequence, structure = Ct.parse(ct_file)
        sequence = standardize_sequence(sequence)

        if sequence_has_regular_characters(sequence):
            return Datapoint(
                sequence=sequence,
                reference=reference,
                structure=structure,
            )

    def from_fasta(sequence, reference, predict_structure):
        """Create a datapoint from a fasta file. The structure and dms will be None."""
        sequence = standardize_sequence(sequence)

        if sequence_has_regular_characters(sequence):
            dotbracket, dms = None, None
            if predict_structure:
                dotbracket = Fasta.predict_structure(sequence)
            return Datapoint(sequence, reference, dotbracket=dotbracket, dms=dms)

    def from_dreem_output(reference, sequence, mutation_rate, predict_structure):
        """Create a datapoint from a dreem output file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        sequence = standardize_sequence(sequence)
        mutation_rate = np.array([float(m) for m in mutation_rate], dtype=np.float32)
        if sequence_has_regular_characters(sequence):
            return Datapoint(
                sequence=sequence,
                reference=reference,
                dotbracket=RNAstructure_singleton.predictStructure(
                    sequence, dms=mutation_rate
                )
                if predict_structure
                else None,
                dms=mutation_rate,
            )

    def from_json_line(ref, d, predict_structure=False):
        """Create a datapoint from a json line. The json line should have the following format:
        "reference": {"sequence": "sequence", "structure": [[1, 2], [3,4]], "dms": [1.0, 2.0, 3.0]}

        Example:
        >>> DatapointFactory.from_json_line("reference", {"sequence": "ACAAGU"})
        Datapoint('reference', sequence='ACAAGU')
        >>> DatapointFactory.from_json_line("reference", {"sequence": "ACAAGU", "structure": [[1, 2], [3, 4]], "dms": [1.0, 2.0, 3.0]})
        Datapoint('reference', sequence='ACAAGU', structure=((1, 2), (3, 4)), dms=(1.0, 2.0, 3.0))
        >>> DatapointFactory.from_json_line("reference", {"sequence": "something else than ACGTUacgtu", "structure": [[1, 2], [3, 4]], "dms": [1.0, 2.0, 3.0]})
        """

        # create the datapoint
        sequence = d["sequence"]
        sequence = standardize_sequence(sequence)

        if predict_structure:
            d["dotbracket"] = RNAstructure_singleton.predictStructure(
                sequence,
                dms=d["dms"] if "dms" in d else None,
                shape=d["shape"] if "shape" in d and "dms" not in d else None,
            )
            if "structure" in d:
                del d["structure"]  # otherwise the dotbracket won't be used

        if sequence_has_regular_characters(sequence):
            return Datapoint(
                reference=ref,
                sequence=sequence,
                structure=d["structure"] if "structure" in d else None,
                dotbracket=d["dotbracket"] if "dotbracket" in d else None,
                dms=d["dms"] if "dms" in d else None,
                shape=d["shape"] if "shape" in d else None,
            )



import os


class Env:
    """Class to get environment variables."""

    def get_hf_token() -> str:
        if "HUGGINGFACE_TOKEN" in os.environ:
            return os.environ["HUGGINGFACE_TOKEN"]
        raise Exception("HUGGINGFACE_TOKEN not found in environment variables")

    def get_rnastructure_path() -> str:
        if "RNASTRUCTURE_PATH" in os.environ:
            return os.environ["RNASTRUCTURE_PATH"]
        return ""

    def get_rnastructure_temp_path() -> str:
        if "RNASTRUCTURE_TEMP_PATH" in os.environ:
            return os.environ["RNASTRUCTURE_TEMP_PATH"]
        return os.path.join(os.path.dirname(__file__), "..", "temp")



import unittest
from rouskinhf import list_datapoints, util
class TestDataDump(unittest.TestCase):
    def test(self):
        listdata = list_datapoints.ListofDatapoints(
            [list_datapoints.Datapoint(sequence='ACACAUCUG', reference='1', dotbracket='.........'),
            list_datapoints.Datapoint(sequence='ACACAUCUU', reference='2', dotbracket='(.......)')]
        )
        listdata.datapoints = list_datapoints.ListofDatapoints(verbose=False).from_pandas(listdata.to_pandas())

        util.dump_json(
            listdata.to_dict(),
            "test.json"
        )


from setuptools import setup

requirements = open('./requirements.txt').read().splitlines()

setup(
    name="rouskinhf",
    packages=["rouskinhf"],
    license="MIT",
    install_requires=requirements,
)